#####################################################
#####################################################
## Licensed Materials - Property of IBM.
## (C) Copyright IBM Corporation 2004, 2017
## All Rights Reserved.
##
## US Government Users Restricted Rights -
## Use, duplication or disclosure restricted by
## GSA ADP Schedule Contract with IBM Corporation.
##
## Filename   : bcast.tcl
##
## Purpose    : TCL event broadcasting package
##
#####################################################
#####################################################

namespace eval BCast {
    variable version            "1.0"
    variable registry
    variable mctmatch
    variable rid                0
    variable listeners
    variable debug              0
}

package provide BCast $BCast::version


####
#
proc BCast::List_Registry { {idlist all} } {
    variable registry
    variable mctmatch

    if {[string equal $idlist "all"]} {
        set idlist [array names registry]
    }

    foreach i $idlist {
        if {[info exists registry($i)]} {
            puts "[list $i [list mct $mctmatch($i) call $registry($i)]]"
        }
    }
}

####
#
proc BCast::List_Listeners { {taglist all} } {
    variable listeners

    if {[string equal $taglist "all"]} {
        set taglist [array names listeners]
    }

    foreach t $taglist {
        if {[info exists listeners($t)]} {
            puts "[list $t $listeners($t)]"
        }
    }
}

####
#
proc BCast::Register {mct receptor} {
    variable registry
    variable mctmatch
    variable rid

    if {[string equal $mct "all"]} {
        set mct "*:*:*"
    }

    set id $rid
    set registry($id) $receptor
    set mctmatch($id) $mct
    incr rid

    return $id
}

####
#
proc BCast::Resign {idlist} {
    variable registry
    variable mctmatch

    BCast::Ignore $idlist
    foreach i $idlist {
        if {[info exists registry($i)]} {
            unset registry($i)
            unset mctmatch($i)
        }
    }
}

####
#
proc BCast::Listen {taglist idlist} {
    variable registry
    variable listeners

    if {[string equal $taglist "all"]} {
        set taglist [array names listeners]
    }

    foreach i $idlist {
        if {[info exists registry($i)]} {
            foreach t $taglist {
                if {![info exists listeners($t)] || ([lsearch listeners($t) $i] < 0)} {
                    lappend listeners($t) $i
                }
            }
        } else {
            return -code error "BCast::Listen -- id '$i' not registered"
        }
    }
}

####
#
proc BCast::Listeners { {taglist all} } {
    variable listeners

    if {[string equal $taglist "all"]} {
        set taglist [array names listeners]
    }

    if {[llength $taglist] > 1} {
        set total 0
        foreach t $taglist {
            set n [llength $listeners($t)]
            set total [expr $total + $n]
            lappend retlist $t $n
        }
        lappend retlist total $total
    } else {
        set retlist [llength $listeners($taglist)]
    }

    return $retlist
}

####
#
proc BCast::Ignore {idlist {taglist all} } {
    variable listeners
    variable registry

    if {[string equal $taglist "all"]} {
        set taglist [array names listeners]
    }

    foreach i $idlist {
        if {[info exists registry($i)]} {
            foreach t $taglist {
                if {[info exists listeners($t)]} {
                    set idx [lsearch $listeners($t) $i]
                    if {$idx >= 0} {
                        set listeners($t) [lreplace $listeners($t) $idx $idx]
                    }
                    if {[llength $listeners($t)] == 0} {
                        unset listeners($t)
                    }
                }
            }
        }
    }
}

####
#
proc BCast::Announce {taglist mct args} {
    variable listeners
    variable registry
    variable mctmatch
    variable debug

    if {$debug} {
        puts "BCast::Announce taglist=$taglist mctlist=$mctlist args=$args"
    }

    if {[string equal $taglist "all"]} {
        set taglist [array names listeners]
    }

    switch -exact -- $mct {
        all {
            foreach t $taglist {
                if {[info exists listeners($t)]} {
                    foreach i $listeners($t) {
                        if {[info exists registry($i)]} {
                            eval $registry($i) mct all $args
                        }
                    }
                }
            }
        }
        default {
            foreach t $taglist {
                if {[info exists listeners($t)]} {
                    foreach i $listeners($t) {
                        if {[info exists registry($i)] && [string match $mctmatch($i) $mct]} {
                            eval $registry($i) mct $mct $args
                        }
                    }
                }
            }
        }
    }
}
