#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2009, 2017
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
# Syscall triggers
#------------------------------------------------------------------------

namespace eval syscall_triggers {
    variable version       "1.0"
    variable verbose       0
    variable fd_table      { { stdin 1 } { stdout 1 } { stderr 1 } }
    variable pipedata
    variable uname
    variable futex_waiters
    array set uname        { sysname Linux nodename localhost release 4.4.0 version Mambo-System machine ppc64 }
    array set futex_waiters {}
}

proc syscall_triggers::copy_buf_to_mem { sim ra buf len } {
    for {set i 0} {$i < $len} {incr i} {
        binary scan [string index $buf $i] "H2" c
        $sim memory set $ra 1 0x$c
        incr ra
    }
}

proc syscall_triggers::copy_string_to_mem { sim ra buf } {
    syscall_triggers::copy_buf_to_mem $sim $ra $buf [string length $buf]
}

# -----------------
# sc_read -- syscall 3
# -----------------
proc syscall_triggers::sc_read { argstring } {
    variable verbose
    variable fd_table
    variable pipedata
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set fd [$sim cpu $cpu display gpr 3]
        set buf_ea [$sim cpu $cpu display gpr 4]
        set len [$sim cpu $cpu display gpr 5]

        if {$verbose} {
            puts "[$sim display cycles]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] read fd=$fd buf=$buf_ea len=$len"
        }

        set buf_ra [$sim cpu $cpu dtranslate $buf_ea]

        array set fd_info [lindex $fd_table $fd]

        if {[info exists fd_info(channel)]} {
            set ch $fd_info(channel)
            set len [string length [set buf [read $ch $len]]]
            copy_buf_to_mem $sim $buf_ra $buf $len
        } elseif {[info exists fd_info(pipe)]} {
            set pipe $fd_info(pipe)
            if {[string length $pipedata($pipe)] == 0} {
                # Need to stall the thread here
                puts "Read from empty pipe not yet implemented!"
                simstop
            }
            if {$len > [string length $pipedata($pipe)]} {
                set len [string length $pipedata($pipe)]
            }
            copy_buf_to_mem $sim $buf_ra $pipedata($pipe) $len
            set pipedata($pipe) [string replace $pipedata($pipe) 0 [expr $len-1]]
        } else {
            puts "sc_read for fd=$fd not implemented."
            set len 0
        }

        # Return value is number of bytes read
        $sim cpu $cpu set gpr 3 $len

        if {$verbose} {
            puts "[$sim display cycles]: $cpu: return from sc_read retval=$len"
        }

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_write -- syscall 4
# -----------------
proc syscall_triggers::sc_write { argstring } {
    variable verbose
    variable fd_table
    variable pipedata
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set fd [$sim cpu $cpu display gpr 3]
        set buf_ea [$sim cpu $cpu display gpr 4]
        set len [$sim cpu $cpu display gpr 5]

        if {$verbose} {
            puts "[$sim display cycles]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] write fd=$fd buf=$buf_ea len=$len"
        }

        set buf_ra [$sim cpu $cpu dtranslate $buf_ea]

        array set fd_info [lindex $fd_table $fd]

        if {[info exists fd_info(stdout)] || [info exists fd_info(stderr)]} {
            puts -nonewline [binary format "c*" [$sim memory display $buf_ra 1 $len]]
        } elseif {[info exists fd_info(channel)]} {
            set ch $fd_info(channel)
            puts -nonewline $ch [binary format "c*" [$sim memory display $buf_ra 1 $len]]
        } elseif {[info exists fd_info(pipe)]} {
            set pipe $fd_info(pipe)
            append pipedata($pipe) [binary format "c*" [$sim memory display $buf_ra 1 $len]]
        } else {
            puts "sc_write for fd=$fd not implemented."
            set len 0
        }

        # Return value is number of bytes written
        $sim cpu $cpu set gpr 3 $len

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_open -- syscall 5
# -----------------
proc syscall_triggers::sc_open { argstring } {
    variable verbose
    variable fd_table
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set path_ea [$sim cpu $cpu display gpr 3]
        set oflags [$sim cpu $cpu display gpr 4]

        set path_ra [$sim cpu $cpu dtranslate $path_ea]
        set path [$sim cpu $cpu memory display $path_ra STRING]

        set flags [lindex {RDONLY WRONLY RDWR} [expr $oflags&0x3]]
        if {$oflags & 0x40}  {lappend flags CREAT}
        if {$oflags & 0x80}  {lappend flags EXCL}
        if {$oflags & 0x100} {lappend flags NOCTTY}
        if {$oflags & 0x200} {lappend flags TRUNC}
        if {$oflags & 0x400} {lappend flags APPEND}

        if {$verbose} {
            puts "[$sim display cycle]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] open path=\"$path\" flags=$flags ($oflags)"
        }

        if {[string compare $path "/dev/tty"] == 0} {
            fconfigure stdout -translation binary -encoding binary
            $sim cpu $cpu set gpr 3 1
        } else {
            set fd [llength $fd_table]
            set ch [open $path $flags]
            lappend fd_table " path $path channel $ch "
            $sim cpu $cpu set gpr 3 $fd
        }

    } errs]} { puts "[info level [info level]] failed: $errs" 
	    $sim cpu $cpu set gpr 3 -1
	}
}

# -----------------
# sc_close -- syscall 6
# -----------------
proc syscall_triggers::sc_close { argstring } {
    variable verbose
    variable fd_table
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set fd [$sim cpu $cpu display gpr 3]

        if {$verbose} {
            puts "[$sim display cycle]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] close fd=$fd"
        }

        array set fd_info [lindex $fd_table $fd]

        if {[info exists fd_info(channel)]} {
            set ch $fd_info(channel)
            close $ch
            array unset fd_info channel
            lset fd_table $fd [array get fd_info]
        } else {
            puts "sc_close for fd=$fd not implemented."
        }

        # Return value of zero means close successful
        $sim cpu $cpu set gpr 3 0

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_time -- syscall 13
# returns the time as the number of seconds since the Epoch, 1970-01-01 00:00:00 +0000 (UTC).
# -----------------
proc syscall_triggers::sc_time { argstring } {
    variable verbose
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        if {$verbose} {
            puts "[$sim display cycle]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] time"
        }

        regexp {^(\d+)} [$sim display rtc] match clockval

        set time_ea [$sim cpu $cpu display gpr 3]

        if {$time_ea != 0} {
            set time_ra [$sim cpu $cpu dtranslate $time_ea]

            # The size of a time_t depends on whether the app is running 32 or 64 bit

            if {[$sim cpu $cpu display spr msr] & 0x8000000000000000} {
                # 64-bit app -- time_t is 8 bytes
                $sim memory set $time_ra 8 $clockval
            } else {
                # 32-bit app -- time_t is 4 bytes
                $sim memory set $time_ra 4 $clockval
            }
        }

        # Return time value
        $sim cpu $cpu set gpr 3 $clockval

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_lseek -- syscall 19
# reposition read/write file offset
# -----------------
proc syscall_triggers::sc_lseek { argstring } {
    variable verbose
    variable fd_table
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set fd [$sim cpu $cpu display gpr 3]
        set offset [$sim cpu $cpu display gpr 4]
        set whence [$sim cpu $cpu display gpr 5]

        if {$verbose} {
            set whence_str [lindex {SEEK_SET SEEK_CUR SEEK_END} $whence]
            puts "[$sim display cycle]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] lseek\
                      fd=$fd offset=$offset whence=$whence_str"
        }

        array set fd_info [lindex $fd_table $fd]

        if {[info exists fd_info(channel)]} {
            set ch $fd_info(channel)
            if {$whence == 0} {
                seek $ch $offset start
            } elseif {$whence == 1} {
                seek $ch $offset current
            } elseif {$whence == 2} {
                seek $ch $offset end
            } else {
                error "Invalid whence value ($whence) in sc_seek"
            }

            # Return value is the new offset for the file
            $sim cpu $cpu set gpr 3 [tell $ch]

        } else {
            puts "sc_seek for fd=$fd not implemented."
        }

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_pipe -- syscall 42
# -----------------
proc syscall_triggers::sc_pipe { argstring } {
    variable verbose
    variable fd_table
    variable pipedata
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set fd_ea [$sim cpu $cpu display gpr 3]
        set fd_ra [$sim cpu $cpu dtranslate $fd_ea]

        set fd0 [llength $fd_table]
        lappend fd_table " pipe $fd0 "
        set pipedata($fd0) ""
        $sim memory set $fd_ra 4 $fd0

        set fd1 [llength $fd_table]
        lappend fd_table " pipe $fd0 "
        $sim memory set [expr $fd_ra+4] 4 $fd1

        # Return value of zero means success
        $sim cpu $cpu set gpr 3 0

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_times -- syscall 43
# get process times
# -----------------
proc syscall_triggers::sc_times { argstring } {
    variable verbose
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set tms_ea [$sim cpu $cpu display gpr 3]

        if {$verbose} {
            puts "[$sim display cycle]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] times"
        }

        # Times are reported in clock ticks, which are 1/100 secs
        regexp {(\d+)([MG])Hz} [$sim cpu $cpu display frequency] match freq units
        if {$units eq "G"} {set freq [expr $freq*1000]}
        set user_ticks [expr [$sim display cycles]/($freq*1000000/100)]

        set tms_ra [$sim cpu $cpu dtranslate $tms_ea]

        # The size of a clock_t depends on whether the app is running 32 or 64 bit

        if {[$sim cpu $cpu display spr msr] & 0x8000000000000000} {

            # 64-bit app -- clock_t is 8 bytes

            $sim memory set [expr $tms_ra+0]  8 $user_ticks   ;# user time
            $sim memory set [expr $tms_ra+8]  8 0             ;# system time
            $sim memory set [expr $tms_ra+16] 8 0             ;# user time of children
            $sim memory set [expr $tms_ra+24] 8 0             ;# system time of children
        } else {

            # 32-bit app -- clock_t is 4 bytes

            $sim memory set [expr $tms_ra+0]  4 $user_ticks   ;# user time
            $sim memory set [expr $tms_ra+4]  4 0             ;# system time
            $sim memory set [expr $tms_ra+8]  4 0             ;# user time of children
            $sim memory set [expr $tms_ra+12] 4 0             ;# system time of children
        }

        # Return value of zero means success
        $sim cpu $cpu set gpr 3 [$sim display cycles]

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_getrusage -- syscall 77
# get resource usage
# -----------------
proc syscall_triggers::sc_getrusage { argstring } {
    variable verbose
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set who [$sim cpu $cpu display gpr 3]
        set usage_ea [$sim cpu $cpu display gpr 4]

        if {$verbose} {
            puts "[$sim display cycle]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] getrusage who=$who"
        }

        regexp {(\d+)([MG])Hz} [$sim cpu $cpu display frequency] match freq units
        if {$units eq "G"} {set freq [expr $freq*1000]}
        set user_secs [expr [$sim display cycles]/($freq*1000000)]
        set user_usecs [expr ([$sim display cycles]%($freq*1000000))/$freq]

        set usage_ra [$sim cpu $cpu dtranslate $usage_ea]

        # The size of a clock_t depends on whether the app is running 32 or 64 bit

        if {[$sim cpu $cpu display spr msr] & 0x8000000000000000} {

            # 64-bit app -- clock_t is 8 bytes

            $sim memory set [expr $usage_ra+0]  8 $user_secs   ;# user secs
            $sim memory set [expr $usage_ra+8]  8 $user_usecs  ;# user usecs
            $sim memory set [expr $usage_ra+16] 8 0            ;# system secs
            $sim memory set [expr $usage_ra+24] 8 0            ;# system usecs
            # For now, don't both setting any addition fields in the usage struct
        } else {

            # 32-bit app -- clock_t is 4 bytes

            $sim memory set [expr $usage_ra+0]  4 $user_secs   ;# user secs
            $sim memory set [expr $usage_ra+4]  4 $user_usecs  ;# user usecs
            $sim memory set [expr $usage_ra+8]  4 0            ;# system secs
            $sim memory set [expr $usage_ra+12] 4 0            ;# system usecs
            # For now, don't both setting any addition fields in the usage struct
        }

        # Return value of zero means success
        $sim cpu $cpu set gpr 3 0

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_mmap -- syscall 90
# -----------------
proc syscall_triggers::sc_mmap { argstring } {
    variable verbose
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set start  [$sim cpu $cpu display gpr 3]
        set size   [$sim cpu $cpu display gpr 4]
        set prot   [$sim cpu $cpu display gpr 5]
        set flags  [$sim cpu $cpu display gpr 6]
        set fd     [$sim cpu $cpu display gpr 7]
        set offset [$sim cpu $cpu display gpr 8]

        if {$verbose} {
            puts "[$sim display cycles]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] sc_mmap\
                      start=[format {0x%x} $start] size=[format {0x%x} $size] prot=[format {0x%x} $start]\
                      flags=[format {0x%x} $flags] fd=[format {%d} $fd] offset=[format {0x%x} $offset]"
        }

        # TODO: fill in the implementation for mmap here

        set retval -1
        set LINUX_MAP_ANONYMOUS 0x20

        if {$flags & $LINUX_MAP_ANONYMOUS} {

            # Anonymous memory - just allocate memory to the process
            set retval [$sim display loader_high_mem]

            # Round up size to be allocated to a 4K boundary
            set alloc_size [expr (($size+0xFFF)/0x1000)*0x1000]

            $sim set loader_high_mem [expr $retval+$alloc_size]
        }

        # Return value is address of allocated memory or -1 on fail
        $sim cpu $cpu set gpr 3 $retval

    } errs]} { puts "[info level [info level]] failed: $errs"; simstop }
}

# -----------------
# sc_fstat -- syscall 108
# -----------------
proc syscall_triggers::sc_fstat { argstring } {
    variable verbose
    variable fd_table
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set fd [$sim cpu $cpu display gpr 3]
        set buf_ea [$sim cpu $cpu display gpr 4]

        if {$verbose} {
            puts "[$sim display cycle]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] fstat fd=$fd buf=$buf_ea"
        }

        set buf_ra [$sim cpu $cpu dtranslate $buf_ea]

        array set fd_info [lindex $fd_table $fd]

        if {[info exists fd_info(stdout)] || [info exists fd_info(stderr)]} {
            set now [clock seconds]
            array set stat_data " dev 6 ino $fd nlink 1 mode 4480 uid 1 gid 1 size 0 \
                                  type characterSpecial atime $now mtime $now ctime $now "
        } elseif {[info exists fd_info(path)]} {
            set path $fd_info(path)
            file stat $path stat_data
        } else {
            puts "sc_fstat for fd=$fd not implemented."
        }

        if { [info exists stat_data] } {
            # Note: the following assumes 64-bit apps

            # Start by clearing out the whole structure
            copy_string_to_mem $sim $buf_ra [string repeat [binary format "c" 0] 144]

            $sim memory set [expr $buf_ra+0]   8 $stat_data(dev)
            $sim memory set [expr $buf_ra+8]   8 $stat_data(ino)
            $sim memory set [expr $buf_ra+16]  8 $stat_data(nlink)
            $sim memory set [expr $buf_ra+24]  4 $stat_data(mode)
            $sim memory set [expr $buf_ra+28]  4 $stat_data(uid)
            $sim memory set [expr $buf_ra+32]  4 $stat_data(gid)
            # leave rdev set to 0
            $sim memory set [expr $buf_ra+48]  8 $stat_data(size)
            $sim memory set [expr $buf_ra+56]  8 1024      ;# blksize
            # leave blocks set to 0
            $sim memory set [expr $buf_ra+72]  8 $stat_data(atime)
            $sim memory set [expr $buf_ra+88]  8 $stat_data(mtime)
            $sim memory set [expr $buf_ra+104] 8 $stat_data(mtime)            
        }

        $sim cpu $cpu set gpr 3 0

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_clone -- syscall 120
# -----------------
proc syscall_triggers::sc_clone { argstring } {
    variable verbose
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set flags [$sim cpu $cpu display gpr 3]
        set stack [$sim cpu $cpu display gpr 4]
        set child_tid [$sim cpu $cpu display gpr 6]

        if {$verbose} {
            puts "[$sim display cycle]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] clone child_stack=$stack flags=$flags"
        }

        set found_cpu 0
        for {set i 0} {$i < [$sim mcm 0 display number_of_cpus]} {incr i} {
            for {set j 0} {$j < [$sim mcm 0 display number_of_threads]} {incr j} {
                set status [$sim cpu $i:$j display status]
                if {[regexp {Initial state} $status]} {
                    set found_cpu 1
                    break
                }
            }
            if ($found_cpu) {break}
        }

        set CLONE_SETTLS 0x00080000

        if {$found_cpu} {
            if {$verbose} {
                puts "Starting new thread on cpu $i:$j"
            }

            # Copy regs from parent to child
            for {set reg 0} {$reg < 32} {incr reg} {
                $sim cpu $i:$j set gpr $reg [$sim cpu $cpu display gpr $reg]
            }
            $sim cpu $i:$j set gpr 1 $stack
            $sim cpu $i:$j set gpr 3 0x0
            if { $flags && $CLONE_SETTLS } {
                $sim cpu $i:$j set gpr 13 $child_tid
            }

            $sim cpu $i:$j set spr pc [expr [$sim cpu $cpu display spr pc]+4]
            $sim cpu $i:$j set spr msr [$sim cpu $cpu display spr msr]

            $sim cpu $i:$j config_on

            # Create a pid to return to the parent
            $sim cpu $cpu set gpr 3 [expr $i+$j+1]
        } else {
            puts "Failed to find available cpu for new thread"
        }

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_uname -- syscall 122
# -----------------
proc syscall_triggers::sc_uname { argstring } {
    variable verbose
    variable uname
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set uts_ea [$sim cpu $cpu display gpr 3]
        set uts_ra [$sim cpu $cpu dtranslate $uts_ea]

        set UTSNAME_LEN 65

        # Start by clearing out the whole structure
        copy_string_to_mem $sim $uts_ra [string repeat [binary format "c" 0] [expr 5*$UTSNAME_LEN]]

        copy_string_to_mem $sim $uts_ra                       $uname(sysname)
        copy_string_to_mem $sim [expr $uts_ra+$UTSNAME_LEN]   $uname(nodename)
        copy_string_to_mem $sim [expr $uts_ra+2*$UTSNAME_LEN] $uname(release)
        copy_string_to_mem $sim [expr $uts_ra+3*$UTSNAME_LEN] $uname(version)
        copy_string_to_mem $sim [expr $uts_ra+4*$UTSNAME_LEN] $uname(machine)

        # Return value of zero means success
        $sim cpu $cpu set gpr 3 0

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

# -----------------
# sc_futex -- syscall 221
# -----------------
proc syscall_triggers::sc_futex { argstring } {
    variable verbose
    variable futex_waiters
    if {[catch {
        array set args $argstring
        foreach var {sim mcm cpu thread sc} {eval set $var $args($var)}
        set cpu $mcm:$cpu:$thread

        set uaddr [$sim cpu $cpu display gpr 3]
        set op [$sim cpu $cpu display gpr 4]
        set val [$sim cpu $cpu display gpr 5]
        set timeout [$sim cpu $cpu display gpr 6]

        if {$verbose} {
            puts "[$sim display cycles]: $cpu: SysCall PC=[$sim cpu $cpu display spr pc] sc_futex\
                      uaddr=[format {0x%x} $uaddr] op=[format {0x%x} $op] val=[format {0x%x} $val]"
        }

        set uaddr_ra [$sim cpu $cpu dtranslate $uaddr]

        set FUTEX_WAIT         0
        set FUTEX_WAKE         1

        if { ($op & 0x7f) == $FUTEX_WAIT } { 

            # This operation atomically verifies that the futex address uaddr still contains the
            # value val, and sleeps awaiting FUTEX_WAKE on this futex address.  If the timeout
            # argument is non-NULL, its contents describe the maximum duration of the wait, which
            # is infinite otherwise.

            if { [$sim memory display $uaddr_ra 4] == $val } {

                if { $verbose } {
                    puts "CPU $cpu blocked on futex=$uaddr"
                    if { $timeout != 0 } {
                        puts "WARNING: sc_futex ignoring timeout on FUTEX_WAIT"
                    }
                }
                if { ! [info exists futex_waiters($uaddr_ra)] } {
                    set futex_waiters($uaddr_ra) [list $cpu]
                } else {
                    lappend futex_waiters($uaddr_ra) $cpu
                }
                $sim cpu $cpu stall
            }

            # claim success
            $sim cpu $cpu set gpr 3 0

        } elseif { ($op & 0x7f) == $FUTEX_WAKE } {

            # This operation wakes at most val processes waiting on this futex address (i.e.,
            # inside FUTEX_WAIT).  Returns the number of processes woken up.

            if { [info exists futex_waiters($uaddr_ra)] && [llength $futex_waiters($uaddr_ra)] > 0 } {

                set cpu_to_wakeup [lindex $futex_waiters($uaddr_ra) 0]
                set futex_waiters($uaddr_ra) [lrange $futex_waiters($uaddr_ra) 1 end]
                if { $verbose } {
                    puts "CPU $cpu_to_wakeup being woken up from wait on futex=$uaddr"
                }

                $sim cpu $cpu_to_wakeup config_on

                # Indicate that one process was woken up.
                $sim cpu $cpu set gpr 3 1
            } else {

                if { $verbose } {
                    puts "No processes to wake up for futex=$uaddr"
                }

                # No processes woken up
                $sim cpu $cpu set gpr 3 0
            }

        } else {
            # No other operations currently implemented
        }

    } errs]} { puts "[info level [info level]] failed: $errs" }
}

proc syscall_triggers::enable { sim args } {
#    $sim trigger set syscall   3 syscall_triggers::sc_read
#    $sim trigger set syscall   4 syscall_triggers::sc_write
#    $sim trigger set syscall   5 syscall_triggers::sc_open
#    $sim trigger set syscall   6 syscall_triggers::sc_close
    $sim trigger set syscall  13 syscall_triggers::sc_time
    $sim trigger set syscall  19 syscall_triggers::sc_lseek
    $sim trigger set syscall  42 syscall_triggers::sc_pipe
    $sim trigger set syscall  43 syscall_triggers::sc_times
    $sim trigger set syscall  77 syscall_triggers::sc_getrusage
#    $sim trigger set syscall  90 syscall_triggers::sc_mmap
#    $sim trigger set syscall 108 syscall_triggers::sc_fstat
    $sim trigger set syscall 120 syscall_triggers::sc_clone
    $sim trigger set syscall 122 syscall_triggers::sc_uname    
    $sim trigger set syscall 221 syscall_triggers::sc_futex
}
